import tensorflow as tf
import numpy as np

from . import math as tk
from .utils import wrap_log_k


@wrap_log_k
def log_bessel_k(v, x):
    return _log_bessel_k(v, x)


def _log_bessel_k(v, x, mask=None, max_iter=30, return_counter=False):
    """
    Digital Library of Mathematical Functions: https://dlmf.nist.gov/10.41
    """

    def cond(target, i, pi, update):
        if mask is not None:
            update &= mask
        return tf.reduce_any(update)

    def body(target, i, pi, update):

        def local_cond(ui, j):
            return j >= 0

        def local_body(ui, j):
            ui = ui * q + factor[i, j]
            return ui, j - 1

        ui = tf.zeros_like(v * x)
        ui, _ = tf.while_loop(local_cond, local_body, (ui, i))
        target = tf.where(update, target + ui / pi, target)
        update &= tk.abs(ui / pi) > tol * tk.abs(target)
        return target, i + 1, pi * p, update

    factor = np.zeros((max_iter, max_iter))
    factor[0, 0] = 1.
    for i in range(max_iter - 1):
        for j in range(i + 1):
            k = i + 2. * j
            factor[i + 1, j + 0] -= factor[i, j] * (0.5 * k + 0.125 / (k + 1.))
            factor[i + 1, j + 1] += factor[i, j] * (0.5 * k + 0.625 / (k + 3.))

    x = tf.convert_to_tensor(x)
    v = tf.convert_to_tensor(v, x.dtype)
    if mask is not None:
        mask = tf.convert_to_tensor(mask, tf.bool)

    tol = tk.epsilon(x.dtype)
    factor = tf.convert_to_tensor(factor, x.dtype)
    p = tk.sqrt(tk.square(v) + tk.square(x))
    q = tk.square(v) / (tk.square(v) + tk.square(x))

    log_const = 0.5 * tk.log(0.5 * tk.pi / p) + v * tk.log((v + p) / x) - p
    target = tf.zeros_like(v * x)
    i = tf.cast(0, tf.int32)
    pi = tf.ones_like(p)
    update = tf.ones_like(v * x, tf.bool)
    init = target, i, pi, update
    target, counter, *_ = tf.while_loop(
        cond, body, init, maximum_iterations=max_iter,
    )
    results = log_const + tk.log(target)

    if return_counter:
        return results, i
    return results



import tensorflow as tf

from . import math as tk
from .utils import wrap_log_k
from .utils import log_bessel_recurrence


@wrap_log_k
def log_bessel_k(v, x):
    v = tk.abs(v)
    n = tk.round(v)
    u = v - n
    log_ku, log_kup1 = _log_bessel_ku(u, x)
    return log_bessel_recurrence(log_ku, log_kup1, u, n, x)[0]


def _log_bessel_ku(u, x, mask=None, max_iter=100, return_counter=False):
    """
    N.M. Temme.
    On the numerical evaluation of the modified Bessel function
    of the third kind.
    Journal of Coumputational Physics, 19, 324-337 (1975).
    """

    def calc_gamma(u):
        factor = [
            +1.8437405873009050, -1.1420226803711680,
            -0.0768528408447867, +0.0065165112670737,
            +0.0012719271366546, +0.0003087090173086,
            -0.0000049717367042, -0.0000034706269649,
            -0.0000000331261198, +0.0000000069437664,
            +0.0000000002423096, +0.0000000000367795,
            -0.0000000000001702, -0.0000000000001356,
            -0.00000000000000149,
        ]
        w = 16. * tk.square(u) - 2.
        coef = [None, None]
        for s in range(2):
            prev, curr = 0., 0.
            for fac in reversed(factor[s + 2::2]):
                prev, curr = curr, w * curr + fac - prev
            coef[s] = 0.5 * (w * curr + factor[s]) - prev
        return coef

    def cond(ki, li, i, ci, pi, qi, fi):
        nonzero_update = tk.abs(ci * fi) > tol * tk.abs(ki)
        if mask is not None:
            nonzero_update &= mask
        return tf.reduce_any(nonzero_update)

    def body(ki, li, i, ci, pi, qi, fi):
        j = i + 1.
        cj = ci * tk.square(0.5 * x) / j
        pj = pi / (j - u)
        qj = qi / (j + u)
        fj = (j * fi + pi + qi) / (tk.square(j) - tk.square(u))
        kj = ki + cj * fj
        lj = li + cj * (pj - j * fj)
        return kj, lj, j, cj, pj, qj, fj

    x = tf.convert_to_tensor(x)
    u = tf.convert_to_tensor(u, x.dtype)
    if mask is not None:
        mask = tf.convert_to_tensor(mask, tf.bool)

    tol = tk.epsilon(x.dtype)
    gp, gm = calc_gamma(u)
    lxh = tk.log(0.5 * x)
    mu = u * lxh

    i = tf.cast(0., x.dtype)
    c0 = tf.ones_like(u * x)
    p0 = 0.5 * tk.exp(-mu) / (gp - u * gm)
    q0 = 0.5 * tk.exp( mu) / (gp + u * gm)
    f0 = (gm * tk.cosh(mu) - gp * lxh * tk.sinhc(mu)) / tk.sinc(u)
    k0 = c0 * f0
    l0 = c0 * (p0 - i * f0)
    init = k0, l0, i, c0, p0, q0, f0

    ku, kn, counter, *_ = tf.while_loop(
        cond, body, init, maximum_iterations=max_iter,
    )
    log_ku, log_kup1 = tk.log(ku), tk.log(kn) - lxh

    if return_counter:
        return log_ku, log_kup1, counter

    return log_ku, log_kup1



import tensorflow as tf

from . import math as tk
from .utils import get_deriv_func, extend, find_zero


def sign_bessel_k(v, x, m=0, n=0, name=None):
    with tf.name_scope(name or 'sign_bessel_k_tk2'):
        dtype = tk.common_dtype([v, x])
        x = tf.convert_to_tensor(x, dtype)
        v = tf.convert_to_tensor(v, dtype)
        m = tf.convert_to_tensor(m, tf.int32)
        n = tf.convert_to_tensor(n, tf.int32)
        return _sign_bessel_k_naive(v, x, m, n)


def log_bessel_k(v, x, m=0, n=0, name=None):
    with tf.name_scope(name or 'log_bessel_k_tk2'):
        dtype = tk.common_dtype([v, x])
        x = tf.convert_to_tensor(x, dtype)
        v = tf.convert_to_tensor(v, dtype)
        m = tf.convert_to_tensor(m, tf.int32)
        n = tf.convert_to_tensor(n, tf.int32)
        return _log_bessel_k_custom_gradient(v, x, m, n)


def slog_bessel_k(v, x, m=0, n=0, name=None):
    with tf.name_scope(name or 'slog_bessel_k_tk2'):
        dtype = tk.common_dtype([v, x])
        x = tf.convert_to_tensor(x, dtype)
        v = tf.convert_to_tensor(v, dtype)
        m = tf.convert_to_tensor(m, tf.int32)
        n = tf.convert_to_tensor(n, tf.int32)
        sign = _sign_bessel_k_naive(v, x, m, n)
        logk = _log_bessel_k_custom_gradient(v, x, m, n)
        return sign, logk


def bessel_ke(v, x, m=0, n=0, name=None):
    with tf.name_scope(name or 'bessel_ke_tk2'):
        dtype = tk.common_dtype([v, x])
        x = tf.convert_to_tensor(x, dtype)
        v = tf.convert_to_tensor(v, dtype)
        m = tf.convert_to_tensor(m, tf.int32)
        n = tf.convert_to_tensor(n, tf.int32)
        sign = _sign_bessel_k_naive(v, x, m, n)
        logk = _log_bessel_k_custom_gradient(v, x, m, n)
        return sign * tk.exp(logk + x)


def bessel_k_ratio(v, x, d=1, m=0, n=0, name=None):
    with tf.name_scope(name or 'bessel_k_ratio_tk2'):
        dtype = tk.common_dtype([v, x])
        x = tf.convert_to_tensor(x, dtype)
        v = tf.convert_to_tensor(v, dtype)
        d = tf.convert_to_tensor(d, dtype)
        m = tf.convert_to_tensor(m, tf.int32)
        n = tf.convert_to_tensor(n, tf.int32)
        signd = _sign_bessel_k_naive(v + d, x, m, n)
        logkd = _log_bessel_k_custom_gradient(v + d, x, m, n)
        sign = _sign_bessel_k_naive(v, x, m, n)
        logk = _log_bessel_k_custom_gradient(v, x, m, n)
        return signd * sign * tf.exp(logkd - logk)


@tf.custom_gradient
def _log_bessel_k_custom_gradient(v, x, m, n):
    def _grad(u):
        sign = _sign_bessel_k_naive(v, x, m, n)
        sign_dv = _sign_bessel_k_naive(v, x, m + 1, n)
        logk_dv = _log_bessel_k_custom_gradient(v, x, m + 1, n)
        sign_dx = _sign_bessel_k_naive(v, x, m, n + 1)
        logk_dx = _log_bessel_k_custom_gradient(v, x, m, n + 1)
        return (
            u * sign_dv * sign * tk.exp(logk_dv - logk),
            u * sign_dx * sign * tk.exp(logk_dx - logk),
            None, None,
        )
    logk = _log_bessel_k_naive(v, x, m, n)
    return logk, _grad


def _sign_bessel_k_naive(v, x, m, n):
    dtype = tk.common_dtype([v, x])
    convv = tf.equal(m % 2, 1) & (v < 0)
    convx = tf.equal(n % 2, 1)
    return tf.where(convv ^ convx, tf.cast(-1, dtype), tf.cast(1, dtype))


def _log_bessel_k_naive(
    v, x, m, n, mask=None, dt0=1., tol=1., max_iter=10, bins=128):

    def func(t):
        out = tf.where(
            tf.equal(m % 2, 0), tk.log_cosh(v * t), tk.log_sinh(v * t),
        )
        out -= x * tk.cosh(t)
        out += tf.where(m > 0, mf * tk.log(t), tf.cast(0, dtype))
        out += tf.where(n > 0, nf * tk.log_cosh(t), tf.cast(0, dtype))
        return out

    def func_mth(t):
        return func(t) - th

    dtype = tk.common_dtype([v, x])
    shape = tf.shape(v * x)

    v = tk.abs(v)
    mf = tf.cast(m, dtype)
    nf = tf.cast(n, dtype)

    if mask is None:
        mask = tf.ones(shape, tf.bool)
    condzero = tf.equal(m % 2, 1) & tf.equal(v, 0)
    condinf = tf.equal(x, 0)
    condnan = x < 0
    mask &= ~condzero & ~condinf & ~condnan

    zero_peak = tf.equal(m, 0) & (tf.square(v) + mf < x)
    zero_inf = m > 0

    eps = tk.epsilon(dtype)
    scale = tf.cast(dt0, dtype)
    deriv = get_deriv_func(func)

    dt = tf.where(~zero_peak & mask, scale, tf.cast(0, dtype))
    ts, te = extend(deriv, tf.zeros(shape, dtype), dt)
    tp = find_zero(deriv, te, ts, tol, max_iter)
    fp = func(tp)

    th = fp + tk.log(eps)
    tpm = tk.minimum(tp - bins * eps, tf.cast(0, dtype))
    t0 = find_zero(func_mth, tf.zeros(shape, dtype), tpm, tol, max_iter)

    tpp = tk.maximum(tp + bins * eps, tp * (1 + bins * eps))
    zero_exists = (func_mth(tpp) > 0) & mask
    dt = tf.where(zero_exists, dt0 * scale, tf.cast(0, dtype))
    ts, te = extend(func_mth, tpp, dt)
    t1 = find_zero(func_mth, te, ts, tol, max_iter)

    t = tf.linspace(t0, t1, 2 * bins + 1, axis=0)[1:-1:2]
    h = (t1 - t0) / bins
    sum_eft = h * tk.sum(tk.exp(func(t) - fp), axis=0)
    out = fp + tk.log(sum_eft)

    out = tf.where(condzero, tf.cast(-tk.inf, dtype), out)
    out = tf.where(condinf, tf.cast(tk.inf, dtype), out)
    out = tf.where(condnan, tf.cast(tk.nan, dtype), out)
    return out



import tensorflow as tf
from . import math as tk

from .utils import wrap_log_k
from .series import _log_bessel_ku as log_ku_small_x
from .cfraction import _log_bessel_ku as log_ku_large_x
from .utils import log_bessel_recurrence
from .asymptotic import _log_bessel_k as log_k_large_v


@wrap_log_k
def log_bessel_k(v, x):
    return _log_bessel_k(v, x)


def _log_bessel_k(v, x, return_counter=False):
    x = tf.convert_to_tensor(x)
    v = tf.convert_to_tensor(v, x.dtype)

    v = tk.abs(v)
    n = tk.round(v)
    u = v - n

    small_v_ = v < 25.0
    small_x_ = x < 1.6 + 0.5 * tk.log(v + 1.)

    small_x =  small_x_ & small_v_ & (x > 0.)
    large_x = ~small_x_ & small_v_
    small_v =  small_v_ & (x > 0.)
    large_v = ~small_v_ & (x > 0.)

    lk0s, lk1s, cs = log_ku_small_x(u, x, small_x, return_counter=True)
    lk0l, lk1l, cl = log_ku_large_x(u, x, large_x, return_counter=True)
    lk0 = tf.where(small_x, lk0s, lk0l)
    lk1 = tf.where(small_x, lk1s, lk1l)
    out_small_v = log_bessel_recurrence(lk0, lk1, u, n, x, small_v)[0]
    out_large_v, cv = log_k_large_v(v, x, large_v, return_counter=True)

    out = tf.cast(tk.nan, x.dtype) # x < 0.
    out = tf.where(tf.equal(x, 0.), tf.cast(tk.inf, x.dtype), out)
    out = tf.where(small_v, out_small_v, out)
    out = tf.where(large_v, out_large_v, out)

    if return_counter:
        return out, (cs, cl, cv)
    return out
